from typing import Callable

import torch
import torch.nn as nn

from torch.nn import functional as F

from deepstruct.graph import LayeredGraph
from deepstruct.sparse import MaskedLinearLayer


class LinearBlockUnit(nn.Module):
    def __init__(self, input_size, output_size, stride=1):
        super(LinearBlockUnit, self).__init__()
        self.act = nn.ReLU()
        # self.conv = nn.Conv2d(1, 10, kernel_size=3, stride=stride, padding=1)
        self.linear = nn.Linear(input_size, output_size)
        self.bn = nn.BatchNorm1d(output_size)

    def forward(self, x):
        out = self.act(x)
        out = self.linear(out)
        out = self.bn(out)
        return out


class FiveWaySparseBlockNet(nn.Module):
    def __init__(self, input_size, output_classes, hidden_block_size=100):
        super(FiveWaySparseBlockNet, self).__init__()

        self.input_blocks = nn.ModuleList()
        for _ in range(5):
            self.input_blocks.append(LinearBlockUnit(input_size, hidden_block_size))

        self.hidden_blocks = nn.ModuleList()
        for _ in range(5):
            self.hidden_blocks.append(
                LinearBlockUnit(hidden_block_size, hidden_block_size)
            )

        self.output_blocks = nn.ModuleList()
        for _ in range(5):
            self.output_blocks.append(
                LinearBlockUnit(hidden_block_size, output_classes)
            )

    def forward(self, x):
        input_to_blocks = {}
        for idx, block_unit in enumerate(self.input_blocks):
            input_to_blocks[idx] = block_unit(x)

        hidden_block_results = {}
        for idx, block_unit in enumerate(self.hidden_blocks):
            hidden_block_results[idx] = block_unit(input_to_blocks[idx])

        output_blocks = []
        for idx, block_unit in enumerate(self.output_blocks):
            output_blocks.append(block_unit(hidden_block_results[idx]))

        return torch.mean(torch.stack(output_blocks), dim=0)


cell_constructor_func: Callable[
    [bool, bool, int, int, int], nn.Module
] = lambda is_input, is_output, in_degree, out_degree, layer: nn.Conv2d(10, 10, 1)
skip_map_fn: Callable[[int, int], nn.Module] = lambda layer: nn.Conv2d(10, 10, 1)
layer_channel_size_func: Callable[[int], int] = lambda layer: 10


class DeepDACellNetwork(nn.Module):
    def __init__(
        self,
        num_classes,
        cell_constructor: cell_constructor_func,
        layer_channel_size: layer_channel_size_func,
        fn_skip_map: skip_map_fn,
        structure: LayeredGraph,
    ):
        super(DeepDACellNetwork, self).__init__()

        self._structure = structure
        assert structure.num_layers > 0
        self._layer_channel_size = layer_channel_size

        self._input_nodes = [n for n in structure.nodes if structure.in_degree(n) == 0]
        self._output_nodes = [
            n for n in structure.nodes if structure.out_degree(n) == 0
        ]
        assert len(self._input_nodes) > 0, "No input nodes in structure: len=%d" % len(
            self.input_nodes
        )
        assert (
            len(self._output_nodes) > 0
        ), "No output nodes in structure: len=%d" % len(self.output_nodes)

        # For each node in our structure, we create a cell based on the given cell constructor
        # The cell constructor has to return a nn.Module which maps [B, C_s, N, M, D] -> [B, C_t, N, M]
        # in which B stands for batch size, C_s for the channel size of the source layer defined by
        # layer_channel_size_func, N and M for width and height of the features, C_t for the channel size of the target
        # layer defined by layer_channel_size_func and D for the outgoing degree of the particular node (which needs to
        # be condensed, e.g. by aggregating on the last dimension)
        self._node_cells = {
            node: cell_constructor(
                is_input=node in self._input_nodes,
                is_output=node in self._output_nodes,
                in_degree=structure.in_degree(node),
                out_degree=structure.out_degree(node),
                layer=structure.get_layer(node),
            )
            for node in self._structure.nodes
        }
        self._nodes = nn.ModuleList(list(self._node_cells.values()))

        # Additionaly maps for skip-connections are created. These map channels from one
        self._node_maps = {}
        self._node_maps_list = []
        for source_layer in self._structure.layers[:-1]:
            self._node_maps[source_layer] = {}
            source_output_channel_size = self._layer_channel_size(source_layer + 1)
            for target_layer in self._structure.layers[source_layer + 1 :]:
                target_channel_size = self._layer_channel_size(target_layer + 1)
                if source_output_channel_size is not target_channel_size:
                    # self._node_maps[source_layer][target_layer] = nn.Conv2d(source_output_channel_size, target_channel_size, kernel_size=1)
                    self._node_maps[source_layer][target_layer] = fn_skip_map(
                        source_output_channel_size, target_channel_size
                    )
                    # print(self._node_maps[source_layer][target_layer])
                    self._node_maps_list.append(
                        self._node_maps[source_layer][target_layer]
                    )
        self._maps = nn.ModuleList(self._node_maps_list)

        # Extract number of out channels from node of --last-- layer (not any output node)
        # A node in the last layer has actually the number of out channels everything is mapped to
        any_output_node = self._structure.get_vertices(self._structure.last_layer)[0]
        num_out_channel = self._node_cells[any_output_node].num_out_channel

        self.last_representation_size = 1000
        self.convlast = nn.Conv2d(
            num_out_channel, self.last_representation_size, kernel_size=1
        )
        self.bnlast = nn.BatchNorm2d(self.last_representation_size)
        self.fc = MaskedLinearLayer(self.last_representation_size, num_classes)

    def _apply_skip_layer_adaption(self, source_node, target_node, input):
        source_layer = self._structure.get_layer(source_node)
        target_layer = self._structure.get_layer(target_node)
        self._layer_channel_size(source_layer)
        self._layer_channel_size(target_layer)

    def forward(self, input):
        # input: [batch_size, channels, N, M] = [B, C_in, N, M]

        outputs = {node: None for node in self._structure.nodes}

        for layer in self._structure.layers:
            for current_node in self._structure.get_vertices(layer):

                if current_node in self._input_nodes:
                    current_input = [input]
                else:
                    # current_input = [ self._node_maps[self._structure.get_layer(u)][layer](outputs[u])
                    #                  if layer in self._node_maps[self._structure.get_layer(u)] else outputs[u] for (u, _) in self._structure.in_edges(current_node)]
                    current_input = []
                    for (source, _) in self._structure.in_edges(current_node):
                        source_layer = self._structure.get_layer(source)
                        if layer in self._node_maps[source_layer]:
                            """print('From source layer', source_layer)
                            print('To target layer', layer)
                            print(self._node_maps[source_layer][layer].in_channels)
                            print(self._node_maps[source_layer][layer].out_channels)
                            print('Source output shape:', outputs[source].shape)
                            print('Mapped source output to new channel', self._node_maps[source_layer][layer].out_channels)"""
                            current_input.append(
                                self._node_maps[source_layer][layer](outputs[source])
                            )
                        else:
                            current_input.append(outputs[source])

                feed = torch.stack(current_input, dim=-1)

                outputs[current_node] = self._node_cells[current_node](feed)

            # Force GPU to possibly free memory as stacks of collected inputs/outputs can grow rapidly
            torch.cuda.empty_cache()

        # Collect all outputs of the output nodes into a list and stack it
        # Use the channel output size of the last layer as targeted outgoing channel size
        # Every output node from previous layers are mapped through a convolution to the next layer
        out_list = []
        for node in self._output_nodes:
            source_layer = self._structure.get_layer(node)
            if (
                source_layer is self._structure.last_layer
                or self._structure.last_layer not in self._node_maps[source_layer]
            ):
                if source_layer is not self._structure.last_layer:
                    print(
                        "Channel size source layer:",
                        self._layer_channel_size(source_layer),
                    )
                    print(
                        "Channel size last layer:",
                        self._layer_channel_size(self._structure.last_layer),
                    )
                    print("Output shape:", outputs[node].shape)
                out_list.append(outputs[node])
            else:
                map_to_last_layer = self._node_maps[source_layer][
                    self._structure.last_layer
                ]
                out_list.append(map_to_last_layer(outputs[node]))
        print([o.shape for o in out_list])
        result = torch.mean(torch.stack(out_list), dim=0)  # [B, C_out, N, M]

        # Classifier
        y = F.relu(result)  # [B, C_out, N, M]
        y = self.convlast(y)  # [B, 1000, N, M]
        y = self.bnlast(y)  # [B, 1000, N, M]
        y = F.adaptive_avg_pool2d(y, (1, 1))  # [B, 1000, 1, 1]
        y = y.view(y.size(0), -1)  # [B, 1000]
        return self.fc(y)  # [B, num_classes]
        # return F.log_softmax(y, dim=1) # [B, num_classes]



import networkx as nx


class LayeredGraph(nx.DiGraph):
    @property
    def first_layer(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def last_layer(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def num_layers(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def first_layer_size(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def last_layer_size(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def layers(self):
        """
        :rtype: list[int]
        """
        raise NotImplementedError()

    def get_layer(self, vertex: int):
        """
        :rtype: int
        """
        raise NotImplementedError()

    def get_vertices(self, layer: int):
        """
        :rtype: list[int]
        """
        raise NotImplementedError()

    def get_layer_size(self, layer: int):
        """
        :rtype: int
        """
        raise NotImplementedError()

    def layer_connected(self, layer_index1: int, layer_index2: int):
        """
        :rtype: bool
        """
        raise NotImplementedError()

    def layer_connection_size(self, layer_index1: int, layer_index2: int):
        """
        :rtype: int
        """
        raise NotImplementedError()


class CachedLayeredGraph(LayeredGraph):
    def __init__(self, **attr):
        super(CachedLayeredGraph, self).__init__(**attr)
        self._has_changed = True
        self._layer_index = None
        self._vertex_by_layer = None

    def add_cycle(self, nodes, **attr):
        super(LayeredGraph, self).add_cycle(nodes, **attr)
        self._has_changed = True

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        super(LayeredGraph, self).add_edge(u_of_edge, v_of_edge, **attr)
        self._has_changed = True

    def add_edges_from(self, ebunch_to_add, **attr):
        super(LayeredGraph, self).add_edges_from(ebunch_to_add, **attr)
        self._has_changed = True

    def add_node(self, node_for_adding, **attr):
        super(LayeredGraph, self).add_node(node_for_adding, **attr)
        self._has_changed = True

    def add_nodes_from(self, nodes_for_adding, **attr):
        super(LayeredGraph, self).add_nodes_from(nodes_for_adding, **attr)
        self._has_changed = True

    def add_path(self, nodes, **attr):
        super(LayeredGraph, self).add_path(nodes, **attr)
        self._has_changed = True

    def add_star(self, nodes, **attr):
        super(LayeredGraph, self).add_star(nodes, **attr)
        self._has_changed = True

    def add_weighted_edges_from(self, ebunch_to_add, weight="weight", **attr):
        super(LayeredGraph, self).add_weighted_edges_from(
            ebunch_to_add, weight="weight", **attr
        )
        self._has_changed = True

    def _get_layer_index(self):
        if (
            self._has_changed
            or self._layer_index is None
            or self._vertex_by_layer is None
        ):
            self._build_layer_index()
            self._has_changed = False

        return self._layer_index, self._vertex_by_layer

    def _layer_by_vertex(self, vertex: int):
        return self._get_layer_index()[0][vertex]

    def _vertices_by_layer(self, layer: int):
        return self._get_layer_index()[1][layer]

    def _build_layer_index(self):
        self._layer_index, self._vertex_by_layer = build_layer_index(self)

    @property
    def first_layer(self):
        """
        :rtype: int
        """
        return self.layers[0]

    @property
    def last_layer(self):
        """
        :rtype: int
        """
        return self.layers[-1]

    @property
    def num_layers(self):
        return len(self.layers)

    @property
    def first_layer_size(self):
        return self.get_layer_size(self.layers[0])

    @property
    def last_layer_size(self):
        return self.get_layer_size(self.layers[-1])

    @property
    def layers(self):
        return [layer for layer in self._get_layer_index()[1]]

    def get_layer(self, vertex: int):
        return self._layer_by_vertex(vertex)

    def get_vertices(self, layer: int):
        return self._vertices_by_layer(layer)

    def get_layer_size(self, layer: int):
        return len(self._vertices_by_layer(layer))

    def layer_connected(self, layer_index1: int, layer_index2: int):
        """
        :rtype: bool
        """
        if layer_index1 is layer_index2:
            raise ValueError(
                "Same layer does not have interconnections, it would be split up."
            )
        if layer_index1 > layer_index2:
            tmp = layer_index2
            layer_index2 = layer_index1
            layer_index1 = tmp

        for source_vertex in self.get_vertices(layer_index1):
            for target_vertex in self.get_vertices(layer_index2):
                if self.has_edge(source_vertex, target_vertex):
                    return True
        return False

    def layer_connection_size(self, layer_index1: int, layer_index2: int):
        """
        :rtype: int
        """
        if layer_index1 is layer_index2:
            raise ValueError(
                "Same layer does not have interconnections, it would be split up."
            )
        if layer_index1 > layer_index2:
            tmp = layer_index2
            layer_index2 = layer_index1
            layer_index1 = tmp

        size = 0
        for source_vertex in self.get_vertices(layer_index1):
            for target_vertex in self.get_vertices(layer_index2):
                if self.has_edge(source_vertex, target_vertex):
                    size += 1
        return size


def build_layer_index(graph: nx.DiGraph, layer_index=None):
    """

    :param graph:
    :type graph igraph.Graph
    :param layer_index:
    :return:
    """
    if layer_index is None:
        layer_index = {}

    recursion_call = {"count": 0}

    def get_layer_index(vertex, graph):
        try:
            vertex = int(vertex)
        except TypeError:
            raise ValueError("You have to pass vertex indices to this function.")
        # print('get_layer_index(%s, roots, graph)' % vertex)
        if vertex is None:
            raise ValueError("Given vertex was none.")
        if vertex not in layer_index:
            recursion_call["count"] += 1
            layer_index[vertex] = (
                max(
                    [
                        get_layer_index(v, graph)
                        for v in nx.algorithms.dag.ancestors(graph, vertex)
                    ]
                    + [-1]
                )
                + 1
            )
        return layer_index[vertex]

    for v in graph:
        get_layer_index(v, graph)

    # print('Recursion call: %s' % recursion_call['count'])

    vertices_by_layer = {}
    for v in layer_index:
        idx = layer_index[v]
        if idx not in vertices_by_layer:
            vertices_by_layer[idx] = []
        vertices_by_layer[idx].append(v)

    return layer_index, vertices_by_layer



import networkx as nx
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.autograd import Variable

from deepstruct.graph import CachedLayeredGraph
from deepstruct.graph import LayeredGraph


class DeepCellDAN(nn.Module):
    def __init__(
        self, num_classes, input_channel_size, cell_constructor, structure: LayeredGraph
    ):
        super(DeepCellDAN, self).__init__()

        self._structure = structure
        assert structure.num_layers > 0
        self._num_classes = num_classes
        self._input_channel_size = input_channel_size

        self._input_nodes = [n for n in structure.nodes if structure.in_degree(n) == 0]
        self._output_nodes = [
            n for n in structure.nodes if structure.out_degree(n) == 0
        ]
        assert len(self._input_nodes) > 0, "No input nodes in structure: len=%d" % len(
            self.input_nodes
        )
        assert (
            len(self._output_nodes) > 0
        ), "No output nodes in structure: len=%d" % len(self.output_nodes)

        # Each cell has to have output channel size of 1
        # The input channel size of a cell is its in_degree
        # Except for input nodes, then the channel size equals the input channel size
        self._node_cells = {
            node: cell_constructor(
                is_input=node in self._input_nodes,
                is_output=node in self._output_nodes,
                in_degree=structure.in_degree(node),
                out_degree=structure.out_degree(node),
                layer=structure.get_layer(node),
                input_channel_size=self._input_channel_size,
            )
            for node in self._structure.nodes
        }
        self._nodes = nn.ModuleList(list(self._node_cells.values()))

        self._output_channel_size = 100
        self._last_convolution = nn.Conv2d(
            len(self._output_nodes), self._output_channel_size, (1, 1)
        )
        self._last_batch_norm = nn.BatchNorm2d(self._output_channel_size)
        self._last_fc = nn.Linear(self._output_channel_size, self._num_classes)

    def forward(self, input):
        # input : [batch_size, input_channel_size, ?, ..], e.g. [100, 1, 28, 28] or [100, 3, 32, 32]

        # x = input.flatten(start_dim=1)
        # x: [batch_size, ?*?*..], e.g. [100, 784] or [100, 3072]

        outputs = {
            input_node: self._node_cells[input_node](input)
            for input_node in self._input_nodes
        }
        for node in nx.topological_sort(self._structure):
            # Skip input nodes, as we already processed them
            if node in self._input_nodes:
                continue

            # Collect all inputs of incoming edges and compute the maximum height and width (last two dimensions) of
            # all possible shapes of those tensors
            inputs_raw = []
            max_h = 0
            max_w = 0
            for (u, v) in self._structure.in_edges(node):
                inputs_raw.append(outputs[u])
                max_h = max(max_h, outputs[u].shape[2])
                max_w = max(max_w, outputs[u].shape[3])

            input_stack = []
            for cell_output in inputs_raw:
                # input can be [batch_size, 1, 20?, 20?] or [batch_size, 1, 30?, 20?]
                # it is then zero padded to [batch_size, 1, max(?), max(?)]
                zero_padded_input = torch.zeros(
                    cell_output.shape[0],
                    cell_output.shape[1],
                    max_h,
                    max_w,
                    device=cell_output.device,
                )
                zero_padded_input[
                    :, :, : cell_output.shape[2], : cell_output.shape[3]
                ] = cell_output
                input_stack.append(zero_padded_input)
            # The current input is then concatenated along the channel dimension to build up to
            # current_input: [batch_size, #in_edges(node), max(?), max(?)]
            current_input = torch.cat(input_stack, 1)

            outputs[node] = self._node_cells[node](current_input)

        # Same trick again: compute the maximum height- and width-dimension of all output tensors of nodes which are
        # output nodes. We have to iterate through all of those nodes.
        max_h = 0
        max_w = 0
        for output_node in self._output_nodes:
            max_h = max(max_h, outputs[output_node].shape[2])
            max_w = max(max_w, outputs[output_node].shape[3])

        # Now we obtained the maximum width and height, so we can zero-pad all tensors to that dimension
        final_output_stack = []
        for output_node in self._output_nodes:
            current_output = outputs[output_node]
            # current_output can be [batch_size, 1, 20?, 20?] or [batch_size, 1, 30?, 20?]
            # it is then zero padded to [batch_size, 1, max(?), max(?)]
            zero_padded = torch.zeros(
                current_output.shape[0],
                current_output.shape[1],
                max_h,
                max_w,
                device=current_output.device,
            )
            zero_padded[
                :, :, : current_output.shape[2], : current_output.shape[3]
            ] = current_output
            final_output_stack.append(zero_padded)
        final_output = torch.cat(
            final_output_stack, 1
        )  # which is now [batch_size, #output_nodes, max(?), max(?)]

        # Now apply a last convolution to have a fixed output channel size
        final_output = self._last_convolution(
            final_output
        )  # convolved down to [batch_size, self._output_channel_size, ?, ?]
        final_output = self._last_batch_norm(
            final_output
        )  # [batch_size, self._output_channel_size, ?, ?]
        # And apply pooling, so we know the exact number of final features for a linear classifier
        final_output = torch.nn.functional.adaptive_avg_pool2d(
            final_output, (1, 1)
        )  # [batch_size, self._output_channel_size, 1, 1]
        final_output = final_output.view(
            final_output.size(0), -1
        )  # [batch_size, self._output_channel_size]
        return self._last_fc(final_output)  # [batch_size, _num_classes]


class MaskableModule(nn.Module):
    def apply_mask(self):
        for layer in maskable_layers(self):
            layer.apply_mask()

    def recompute_mask(self, theta=0.0001):
        for layer in maskable_layers(self):
            layer.recompute_mask(theta)


class MaskedDeepDAN(MaskableModule):
    """
    A deep directed acyclic network model which is capable of masked layers and masked skip-layer connections.
    """

    def __init__(self, input_size, num_classes, structure: LayeredGraph):
        super(MaskedDeepDAN, self).__init__()

        self._structure = structure
        assert structure.num_layers > 0

        # Multiple dimensions for input size are flattened out
        if type(input_size) is tuple or type(input_size) is torch.Size:
            input_size = np.prod(input_size)
        input_size = int(input_size)

        self.layer_first = MaskedLinearLayer(input_size, structure.first_layer_size)
        self.layers_main_hidden = nn.ModuleList(
            [
                MaskedLinearLayer(
                    structure.get_layer_size(cur_lay - 1),
                    structure.get_layer_size(cur_lay),
                )
                for cur_lay in structure.layers[1:]
            ]
        )

        for layer_idx, layer in zip(structure.layers[1:], self.layers_main_hidden):
            mask = torch.zeros(
                structure.get_layer_size(layer_idx),
                structure.get_layer_size(layer_idx - 1),
            )
            for source_idx, source_vertex in enumerate(
                structure.get_vertices(layer_idx - 1)
            ):
                for target_idx, target_vertex in enumerate(
                    structure.get_vertices(layer_idx)
                ):
                    if structure.has_edge(source_vertex, target_vertex):
                        mask[target_idx][source_idx] = 1
            layer.set_mask(mask)

        skip_layers = []
        self._skip_targets = {}
        for target_layer in structure.layers[2:]:
            target_size = structure.get_layer_size(target_layer)
            for distant_source_layer in structure.layers[: target_layer - 1]:
                if structure.layer_connected(distant_source_layer, target_layer):
                    if target_layer not in self._skip_targets:
                        self._skip_targets[target_layer] = []

                    skip_layer = MaskedLinearLayer(
                        structure.get_layer_size(distant_source_layer), target_size
                    )
                    mask = torch.zeros(
                        structure.get_layer_size(target_layer),
                        structure.get_layer_size(distant_source_layer),
                    )
                    for source_idx, source_vertex in enumerate(
                        structure.get_vertices(distant_source_layer)
                    ):
                        for target_idx, target_vertex in enumerate(
                            structure.get_vertices(target_layer)
                        ):
                            if structure.has_edge(source_vertex, target_vertex):
                                mask[target_idx][source_idx] = 1
                    skip_layer.set_mask(mask)

                    skip_layers.append(skip_layer)
                    self._skip_targets[target_layer].append(
                        {"layer": skip_layer, "source": distant_source_layer}
                    )
        self.layers_skip_hidden = nn.ModuleList(skip_layers)

        self.layer_out = MaskedLinearLayer(structure.last_layer_size, num_classes)
        self.activation = nn.ReLU()

    def generate_structure(self, include_input=False, include_output=False):
        structure = CachedLayeredGraph()

        node_number_offset = 0
        layer_nodeidx2node = {}

        if include_input:
            layer = self.layer_first
            layer_nodeidx2node[0] = {
                node_idx: node
                for node_idx, node in enumerate(
                    np.arange(
                        node_number_offset, node_number_offset + layer.mask.shape[1]
                    )
                )
            }
            node_number_offset += layer.mask.shape[1]
            structure.add_nodes_from(layer_nodeidx2node[0].values())

            for source_node_idx in range(layer.mask.shape[1]):
                source_node = layer_nodeidx2node[0][source_node_idx]

                layer_nodeidx2node[1] = {
                    node_idx: node
                    for node_idx, node in enumerate(
                        np.arange(
                            node_number_offset, node_number_offset + layer.mask.shape[0]
                        )
                    )
                }
                node_number_offset += layer.mask.shape[0]
                structure.add_nodes_from(layer_nodeidx2node[1].values())

                for target_node_idx in range(layer.mask.shape[0]):
                    if layer.mask[target_node_idx][source_node_idx]:
                        target_node = layer_nodeidx2node[1][target_node_idx]
                        structure.add_edge(source_node, target_node)

        # Main hidden layers
        for source_layer_idx, layer in enumerate(self.layers_main_hidden, start=1):
            if source_layer_idx not in layer_nodeidx2node:
                layer_nodeidx2node[source_layer_idx] = {
                    node_idx: node
                    for node_idx, node in enumerate(
                        np.arange(
                            node_number_offset, node_number_offset + layer.mask.shape[1]
                        )
                    )
                }
                node_number_offset += layer.mask.shape[1]
                structure.add_nodes_from(layer_nodeidx2node[source_layer_idx].values())

            target_layer_idx = source_layer_idx + 1
            if target_layer_idx not in layer_nodeidx2node:
                layer_nodeidx2node[target_layer_idx] = {
                    node_idx: node
                    for node_idx, node in enumerate(
                        np.arange(
                            node_number_offset, node_number_offset + layer.mask.shape[0]
                        )
                    )
                }
                node_number_offset += layer.mask.shape[0]
                structure.add_nodes_from(layer_nodeidx2node[target_layer_idx].values())

            for source_node_idx in range(layer.mask.shape[1]):
                source_node = layer_nodeidx2node[source_layer_idx][source_node_idx]

                for target_node_idx in range(layer.mask.shape[0]):
                    target_node = layer_nodeidx2node[target_layer_idx][target_node_idx]

                    if layer.mask[target_node_idx][source_node_idx]:
                        structure.add_edge(source_node, target_node)

        # Skip layers
        for target_idx in self._skip_targets:
            for target in self._skip_targets[target_idx]:
                source_layer = target["layer"]
                # We have to shift target layer index as layer_nodeidx2node is indexed differently
                target_layer_idx = target_idx + 1
                # Also shift the source layer index by one as we indexed layer_nodeidx2node differently
                source_layer_idx = target["source"] + 1

                if source_layer_idx not in layer_nodeidx2node:
                    # Possibly the source layer (e.g. input layer) was left out, so skip connections from it can not be
                    # considered
                    continue

                source_idx2node = layer_nodeidx2node[source_layer_idx]
                target_idx2node = layer_nodeidx2node[target_layer_idx]

                for source_node_idx in range(source_layer.mask.shape[1]):
                    source_node = source_idx2node[source_node_idx]
                    for target_node_idx in range(source_layer.mask.shape[0]):
                        if source_layer.mask[target_node_idx][source_node_idx]:
                            target_node = target_idx2node[target_node_idx]
                            structure.add_edge(source_node, target_node)

        if include_output:
            layer = self.layer_out
            layer_indices = len(layer_nodeidx2node)
            layer_nodeidx2node[layer_indices] = {
                node_idx: node
                for node_idx, node in enumerate(
                    np.arange(
                        node_number_offset, node_number_offset + layer.mask.shape[0]
                    )
                )
            }
            node_number_offset += layer.mask.shape[0]
            structure.add_nodes_from(layer_nodeidx2node[layer_indices].values())

            for source_node_idx in range(layer.mask.shape[1]):
                source_node = layer_nodeidx2node[layer_indices - 1][source_node_idx]
                for target_node_idx in range(layer.mask.shape[0]):
                    if layer.mask[target_node_idx][source_node_idx]:
                        target_node = layer_nodeidx2node[layer_indices][target_node_idx]
                        structure.add_edge(source_node, target_node)

        return structure

    def forward(self, input):
        # input : [batch_size, ?, ?, ..], e.g. [100, 1, 28, 28] or [100, 3, 32, 32]
        x = input.flatten(start_dim=1)
        # x: [batch_size, ?*?*..], e.g. [100, 784] or [100, 3072]
        last_output = self.activation(self.layer_first(x))
        layer_results = {0: last_output}
        for layer, layer_idx in zip(
            self.layers_main_hidden, self._structure.layers[1:]
        ):
            out = self.activation(layer(last_output))

            if layer_idx in self._skip_targets:
                for skip_target in self._skip_targets[layer_idx]:
                    source_layer = skip_target["layer"]
                    source_idx = skip_target["source"]

                    out += self.activation(source_layer(layer_results[source_idx]))

            layer_results[layer_idx] = out  # copy?
            last_output = out

        return self.layer_out(last_output)


class MaskedDeepFFN(MaskableModule):
    """
    A deep feed-forward network model which is capable of masked layers.
    Masked layers can represent sparse structures between consecutive layers.
    This representation is suitable for feed-forward sparse networks, probably with density 0.5 and above per layer.
    """

    def __init__(self, input_size, num_classes, hidden_layers: list):
        super(MaskedDeepFFN, self).__init__()
        assert len(hidden_layers) > 0

        # Multiple dimensions for input size are flattened out
        if type(input_size) is tuple or type(input_size) is torch.Size:
            input_size = np.prod(input_size)
        input_size = int(input_size)

        self.layer_first = MaskedLinearLayer(input_size, hidden_layers[0])
        self.layers_hidden = nn.ModuleList(
            [
                MaskedLinearLayer(hidden_layers[lay], hid)
                for lay, hid in enumerate(hidden_layers[1:])
            ]
        )
        self.layer_out = MaskedLinearLayer(hidden_layers[-1], num_classes)
        self.activation = nn.ReLU()

    def generate_structure(self, include_input=False, include_output=False):
        """

        :param include_input:
        :param include_output:
        :rtype : LayeredGraph
        :return:
        """
        structure = CachedLayeredGraph()

        def add_edges(structure, layer, offset_source):
            offset_target = offset_source + layer.mask.shape[1]
            for source_node_idx, source_node in enumerate(range(layer.mask.shape[1])):
                for target_node_idx, target_node in enumerate(
                    range(layer.mask.shape[0])
                ):
                    if layer.mask[target_node_idx][source_node_idx]:
                        structure.add_edge(
                            offset_source + source_node, offset_target + target_node
                        )
            return offset_target

        offset_source = 0
        if include_input:
            offset_source = add_edges(structure, self.layer_first, 0)

        for layer in self.layers_hidden:
            offset_source = add_edges(structure, layer, offset_source)

        if include_output:
            add_edges(structure, self.layer_out, offset_source)

        return structure

    def forward(self, input):
        # input : [batch_size, ?, ?, ..], e.g. [100, 1, 28, 28] or [100, 3, 32, 32]
        out = self.activation(
            self.layer_first(input.flatten(start_dim=1))
        )  # [B, n_hidden_1]
        for layer in self.layers_hidden:
            out = self.activation(layer(out))
        return self.layer_out(out)  # [B, n_out]

    def to(self, *args, **kwargs):
        super().to(*args, **kwargs)
        self.layer_first.to(*args, **kwargs)
        for idx, h in enumerate(self.layers_hidden):
            self.layers_hidden[idx] = self.layers_hidden[idx].to(*args, **kwargs)
        self.layer_out.to(*args, **kwargs)
        return self


def maskable_layers(network):
    for child in network.children():
        if type(child) is MaskedLinearLayer:
            yield child
        elif type(child) is nn.ModuleList:
            for layer in maskable_layers(child):
                yield layer


def maskable_layers_with_name(network):
    for name, child in network.named_children():
        if type(child) is MaskedLinearLayer:
            yield name, child
        elif type(child) is nn.ModuleList:
            for name, layer in maskable_layers_with_name(child):
                yield name, layer


def prunable_layers(network):
    return maskable_layers(network)


def prunable_layers_with_name(network):
    return maskable_layers_with_name(network)


class MaskedLinearLayer(nn.Linear):
    def __init__(self, in_feature, out_features, bias=True, keep_layer_input=False):
        """
        :param in_feature:          The number of features that are inserted in the layer.
        :param out_features:        The number of features that are returned by the layer.
        :param bias:                Iff each neuron in the layer should have a bias unit as well.
        """
        super().__init__(in_feature, out_features, bias)

        self.register_buffer(
            "mask", torch.ones((out_features, in_feature), dtype=torch.bool)
        )
        self.keep_layer_input = keep_layer_input
        self.layer_input = None

    def get_mask(self):
        return self.mask

    def set_mask(self, mask):
        self.mask = Variable(mask)

    def get_weight_count(self):
        return self.mask.sum()

    def apply_mask(self):
        # Assigning "self.weight = torch.nn.Parameter(self.weight.mul(self.mask))" might have side effects
        # Using direct manipulation on tensor "self.weight.data"
        self.weight.data = self.weight * self.mask

    def recompute_mask(self, theta: float = 0.001):
        """
        Recomputes the mask based on the weight magnitudes.
        If you want to consider the existing mask, make sure to first call apply_mask() and then recompute it.

        :param theta: Specifies a possible threshold for absolute distance to zero.
        """
        self.mask = torch.ones(
            self.weight.shape, dtype=torch.bool, device=self.mask.device
        )
        self.mask[torch.where(abs(self.weight) < theta)] = False

    def get_weight(self):
        return self.weight

    def reset_parameters(self, keep_mask=False):
        super().reset_parameters()
        # hasattr() is necessary because reset_parameters() is called in __init__ of Linear(), but buffer 'mask'
        # may only be registered after super() call, thus 'mask' might not be defined as buffer / attribute, yet
        if hasattr(self, "mask") and not keep_mask:
            self.mask = torch.ones(self.weight.size(), dtype=torch.bool)

    def forward(self, input):
        x = (
            input.float()
        )  # In case we get a double-tensor passed, force it to be float for multiplications to work

        # Possibly store the layer input
        if self.keep_layer_input:
            self.layer_input = x.data

        return F.linear(x, self.weight * self.mask, self.bias)



import torch


def run_evaluation(test_loader, model, device):
    """
    Test the model on the tests data set provided by the tests loader.

    :param test_loader:     The provided tests data set
    :param model:           The model that should be tested
    :return: The percentage of correctly classified samples from the data set.
    """
    model.eval()

    with torch.no_grad():
        correct = 0
        total = 0
        for test_features, test_labels in test_loader:
            test_features = test_features.to(device)
            test_labels = test_labels.to(device)

            outputs = model(test_features)
            _, predicted = torch.max(outputs.data, 1)
            total += test_labels.size(0)
            correct += (predicted == test_labels).sum().item()
        return 100 * correct / total


def train(train_loader, model, optimizer, criterion, device):
    """
    Train the model on the train data set with a loss function and and optimization algorithm.

    :param train_loader:    The training data set.
    :param model:           The to be trained model.
    :param optimizer:       The used optimizer for the learning process.
    :param criterion:       The loss function of the network.
    :param percentage:      If the function should also calculate the percentage of right made decisions.
    :return: The average loss of the network in this epoch.
    """
    model.train()
    total_loss = 0
    correct = 0
    total = 0

    for i, (features, labels) in enumerate(train_loader):
        features = features.to(device)
        labels = labels.to(device)

        optimizer.zero_grad()

        outputs = model(features)
        loss = criterion(outputs, labels)
        total_loss += loss.item()

        _, predicted = torch.max(outputs.data, 1)
        total += labels.size(0)
        correct += (predicted == labels).sum().item()

        loss.backward()
        optimizer.step()

    return total_loss / len(train_loader), 100 * correct / total



import math
import os

import numpy as np

import deepstruct.sparse


def generate_hessian_inverse_fc(layer, hessian_inverse_path, layer_input_train_dir):
    """
    This function calculate hessian inverse for a fully-connect layer
    :param hessian_inverse_path: the hessian inverse path you store
    :param layer: the layer weights
    :param layer_input_train_dir: layer inputs in the dir
    :return:
    """

    w_layer = layer.get_weight().data.numpy().T
    n_hidden_1 = w_layer.shape[0]

    # Here we use a recursive way to calculate hessian inverse
    hessian_inverse = 1000000 * np.eye(n_hidden_1)

    dataset_size = 0
    for input_index, input_file in enumerate(os.listdir(layer_input_train_dir)):
        layer2_input_train = np.load(layer_input_train_dir + "/" + input_file)

        if input_index == 0:
            dataset_size = layer2_input_train.shape[0] * len(
                os.listdir(layer_input_train_dir)
            )

        for i in range(layer2_input_train.shape[0]):
            # vect_w_b = np.vstack((np.array([layer2_input_train[i]]).T, np.array([[1.0]])))
            vect_w = np.array([layer2_input_train[i]]).T
            denominator = dataset_size + np.dot(
                np.dot(vect_w.T, hessian_inverse), vect_w
            )
            numerator = np.dot(
                np.dot(hessian_inverse, vect_w), np.dot(vect_w.T, hessian_inverse)
            )
            hessian_inverse = hessian_inverse - numerator * (1.00 / denominator)

    np.save(hessian_inverse_path, hessian_inverse)


def get_filtered_saliency(saliency, mask):
    s = list(saliency)
    m = list(mask)

    _, filtered_w = zip(
        *(
            (masked_val, weight_val)
            for masked_val, weight_val in zip(m, s)
            if masked_val == 1
        )
    )
    return filtered_w


def get_layer_count(network):
    i = 0
    for _ in deepstruct.sparse.prunable_layers_with_name(network):
        i += 1
    return i


def get_weight_distribution(network):
    all_weights = []
    for layer in deepstruct.sparse.prunable_layers(network):
        mask = list(layer.get_mask().numpy().flatten())
        weights = list(layer.get_weight().data.numpy().flatten())

        masked_val, filtered_weights = zip(
            *(
                (masked_val, weight_val)
                for masked_val, weight_val in zip(mask, weights)
                if masked_val == 1
            )
        )

        all_weights += list(filtered_weights)

    # return all the weights, that are not masked as a numpy array
    return np.array(all_weights)


def entropy(labels, base=None):
    value, counts = np.unique(labels, return_counts=True)
    norm_counts = counts / counts.sum()
    base = math.e if base is None else base
    return -(norm_counts * np.log(norm_counts) / np.log(base)).sum()


def kullback_leibler(a, b):
    a = np.asarray(a, dtype=np.float)
    b = np.asarray(b, dtype=np.float)

    return np.sum(np.where(a != 0, a * np.log(a / b), 0))



import itertools

import numpy as np
import torch
import torch.nn as nn

from deepstruct.graph import LayeredGraph
from deepstruct.sparse import MaskedDeepDAN
from deepstruct.sparse import MaskedDeepFFN
from deepstruct.util import kullback_leibler


class ScalableFamily(object):
    _precision: int = 2
    _epsilon: float = 0.001
    _entropy_equivalence_epsilon: float = 0.1

    def __init__(self, proportions: np.ndarray):
        assert proportions is not None
        assert len(proportions) > 0
        assert np.abs(np.sum(proportions) - 1) < self._epsilon

        self._proportions = proportions

    @property
    def entropy_similarity_epsilon(self) -> float:
        return self._entropy_equivalence_epsilon

    @entropy_similarity_epsilon.setter
    def entropy_similarity_epsilon(self, epsilon):
        assert epsilon > 0
        self._entropy_equivalence_epsilon = epsilon

    @property
    def precision(self) -> int:
        return self._precision

    @precision.setter
    def precision(self, precision: int):
        assert precision > 0
        self._precision = precision

    @property
    def proportions(self):
        return np.round(self._proportions, self.precision)

    def draw(self, scale: int) -> list:
        return [round(int(np.maximum(1, size))) for size in scale * self.proportions]

    def build(self, input_shape, output_shape, scale: int) -> nn.Module:
        assert scale > 0

        layers = self.draw(scale)
        print("l", layers, np.sum(layers))
        return MaskedDeepFFN(input_shape, output_shape, layers)

    def __eq__(self, other):
        if not isinstance(other, ScalableFamily):
            return False

        return self.entropy_similarity(other) < self.entropy_similarity_epsilon

    def entropy_similarity(self, other) -> float:
        assert isinstance(other, ScalableFamily)
        max_length = np.maximum(len(self._proportions), len(other._proportions))
        return kullback_leibler(
            np.pad(self._proportions, (0, max_length - len(self._proportions))),
            np.pad(other._proportions, (0, max_length - len(other._proportions))),
        )

    def __str__(self):
        return str(self.draw(np.power(10, self.precision)))


class ScalableDAN(object):
    _structure: LayeredGraph
    _proportion_map: dict

    def __init__(self, structure: LayeredGraph, proportion_map: dict):
        if proportion_map is None:
            proportion_map = {v: 1 / len(structure.nodes) for v in structure.nodes}
        self.structure = structure
        self.proportions = proportion_map

    @property
    def structure(self):
        return self._structure

    @structure.setter
    def structure(self, structure):
        assert structure is not None
        assert len(structure.nodes) > 0

        self._structure = structure

    @property
    def proportions(self):
        return self._proportion_map

    @proportions.setter
    def proportions(self, map):
        assert len(map) == len(self.structure.nodes)
        assert np.isclose(sum(map.values()), 1)
        self._proportion_map = map

    def grow(self, scale: int) -> LayeredGraph:
        assert scale > 0

        graph_scaled = CachedLayeredGraph()

        num_nodes = 0
        vertex_correspondences = {}
        for layer in self.structure.layers:
            for v in self.structure.get_vertices(layer):
                size = int(np.round(self.proportions[v] * scale))
                vertex_correspondences[v] = np.arange(num_nodes, num_nodes + size)
                graph_scaled.add_nodes_from(
                    vertex_correspondences[v]
                )  # ['v%s_%s' % (v, idx) for idx in range(size)])
                num_nodes += size

                for (source_vertex, _) in self.structure.in_edges(v):
                    # print(source_vertex, _)
                    # print('nodes source:', vertex_correspondences[source_vertex])
                    # print('nodes target:', vertex_correspondences[v])
                    # print(list(itertools.product(vertex_correspondences[source_vertex], vertex_correspondences[v])))
                    graph_scaled.add_edges_from(
                        itertools.product(
                            vertex_correspondences[source_vertex],
                            vertex_correspondences[v],
                        )
                    )

        return graph_scaled

    def build(self, input_shape, output_shape, scale: int) -> nn.Module:
        assert scale > 0
        graph_scaled = self.grow(scale)
        return MaskedDeepDAN(input_shape, output_shape, graph_scaled)


if __name__ == "__main__":
    from deepstruct.graph import CachedLayeredGraph

    graph = CachedLayeredGraph()
    graph.add_edge(0, 3)
    graph.add_edge(0, 5)
    graph.add_edge(1, 3)
    graph.add_edge(1, 4)
    graph.add_edge(1, 5)
    graph.add_edge(2, 3)
    graph.add_edge(2, 4)

    graph.add_edge(3, 5)
    graph.add_edge(3, 6)
    graph.add_edge(4, 5)
    graph.add_edge(4, 7)

    graph.add_edge(5, 7)

    props = {
        v: p
        for v, p in zip(
            graph.nodes, np.random.dirichlet(np.ones(len(graph.nodes)) * 100)
        )
    }
    print(props)
    fam = ScalableDAN(graph, props)
    model = fam.build(8, 4, 1000)
    print(model)
    for lay in model.layers_main_hidden:
        print(lay)
        print(torch.sum(lay.mask) / np.prod(lay.mask.shape))

    for lay in model.layers_skip_hidden:
        print(lay)
        print(torch.sum(lay.mask) / np.prod(lay.mask.shape))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deepstruct']

package_data = \
{'': ['*']}

install_requires = \
['importlib-metadata>=1.5.0,<2.0.0',
 'importlib-resources>=1.4.0,<2.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'networkx>=2.0',
 'semantic_version>=2.8.4,<3.0.0',
 'torch>=1.4']

setup_kwargs = {
    'name': 'deepstruct',
    'version': '0.5.0',
    'description': '',
    'long_description': '# deepstruct - tools for neural network graph topology analysis ![Tests](https://github.com/innvariant/pypaddle/workflows/Tests/badge.svg)  [![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://GitHub.com/Naereen/StrapDown.js/graphs/commit-activity) [![Python 3.6](https://img.shields.io/badge/python-3.6-blue.svg)](https://www.python.org/downloads/release/python-360/) [![Python 3.7](https://img.shields.io/badge/python-3.7-blue.svg)](https://www.python.org/downloads/release/python-370/) [![Python 3.6](https://img.shields.io/badge/python-3.8-blue.svg)](https://www.python.org/downloads/release/python-380/)\nTools for fusing machine learning and graph theory.\nWe are fascinated with the interplay of end-to-end learnable, locally restricted models and their graph theoretical properties.\nSearching for evidence of the structural prior hypothesis.\nInterested in pruning, neural architecture search or learning theory in general?\n\n**Note** we are currently renaming *pypaddle* to *deepstruct* due to naming collisions and better semantical wording.\n\n## Installation\nVia *poetry* (**recommended** for projects) using PyPi:\n``poetry add deepstruct``\n\nDirectly with *pip* from PyPi:\n```bash\npip install deepstruct\n```\n\nVia *conda* in your *environment.yml* (recommended for reproducible experiments):\n```yaml\nname: exp01\nchannels:\n- defaults\ndependencies:\n- pip>=20\n- pip:\n    - deepstruct\n```\n\nFrom public GitHub:\n```bash\npip install --upgrade git+ssh://git@github.com:innvariant/pypaddle.git\n```\n\n## Models\n- *pypaddle.sparse.MaskableModule*: pytorch modules that contain explicit masks to enforce (mostly zero-ordered) structure\n- *pypaddle.sparse.MaskedLinearLayer*: pytorch module with a simple linear layer extended with masking capability.\nSuitable if you want to have linear-layers on which to enforce masks which could be obtained through pruning, regularization or other other search techniques.\n- *pypaddle.sparse.MaskedDeepFFN*: feed-forward neural network with any width and depth and easy-to-use masks.\nSuitable for simple and canonical pruning research on zero-ordered structure\n- *pypaddle.sparse.MaskedDeepDAN*: feed-forward neural network with skip-layer connections based on any directed acyclic network.\nSuitable for arbitrary structures on zero-order and on that level most flexible but also computationally expensive.\n- *pypaddle.sparse.DeepCellDAN*: complex module based on a directed acyclic network and custom cells on third-order structures.\nSuitable for large-scale neural architecture search\n\n## What is the orders of structure?\n- zero-th order: weight-level\n- first order: kernel-level (filter, channel, blocks, cells)\n- second order: layers\n\nThere is various evidence across empirical machine learning studies that the way artificial neural networks are structurally connected has a (minor?) influence on performance metrics such as the accuracy or probably even on more complex concepts such as adversarial robustness.\nWhat do we mean by "structure"?\nWe define structure over graph theoretic properties given a computational graph with very restricted non-linearities.\nThis includes all major neural network definitions and lets us study them from the perspective of their *representation* and their *structure*.\nIn a probabilistic sense, one can interprete structure as a prior to the model and despite single-layered wide networks are universal function approximators we follow the hypothesis that given certain structural priors we can find models with better properties.\n\n\n\n## Sparse Neural Network implementations\n![Sparse Network Connectivity on zeroth order with a masked deep feed-forward neural network](doc/masked-deep-ffn.png)\n![Sparse Network Connectivity on zeroth order with a masked deep neural network with skip-layer connections](doc/masked-deep-dan.png)\n![Sparse Network Connectivity on second order with a masked deep cell-based neural network](doc/masked-deep-cell-dan.png)\n\n**What\'s contained in pypaddle?**\n- ready-to-use models in pytorch for learning instances on common (supervised/unsupervised) datasets from which a structural analysis is possible\n- model-to-graph transformations for studying models from a graph-theoretic perspective\n\nBefore considering implementations, one should have a look on possible representations of Sparse Neural Networks.\nIn case of feed-forward neural networks (FFNs) the network can be represented as a list of weight matrices.\nEach weight matrix represents the connections from one layer to the next.\nHaving a network without some connections then simply means setting entries in those matrices to zero.\nRemoving a particular neuron means setting all entries representing its incoming connections to zero.\n\nHowever, sparsity can be employed on various levels of a general artificial neural network.\nZero order sparsity would remove single weights (representing connections) from the network.\nFirst order sparsity removes groups of weights within one dimension of a matrix from the network.\nSparsity can be employed on connection-, weight-, block-, channel-, cell-level and so on.\nImplementations respecting the areas for sparsification can have drastical differences.\nThus there are various ways for implementing Sparse Neural Networks.\n\n### Feed-forward Neural Network with sparsity\nThe simplest implementation is probably one which provides multiple layers with binary masks for each weight matrix.\nIt doesn\'t consider any skip-layer connections.\nEach layer is then connected to only the following one.\n```python\nimport pypaddle.sparse\n\nmnist_model = pypaddle.sparse.MaskedDeepFFN((1, 28, 28), 10, [100, 100])\n```\n\n\n```python\nimport pypaddle.sparse\n\nstructure  = pypaddle.sparse.CachedLayeredGraph()\n# .. add nodes & edges to the networkx graph structure\n\n# Build a neural network classifier with 784 input and 10 output neurons and the given structure\nmodel = pypaddle.sparse.MaskedDeepDAN(784, 10, structure)\nmodel.apply_mask()  # Apply the mask on the weights (hard, not undoable)\nmodel.recompute_mask()  # Use weight magnitude to recompute the mask from the network\npruned_structure = model.generate_structure()  # Get the structure -- a networkx graph -- based on the current mask\n\nnew_model = pypaddle.sparse.MaskedDeepDAN(784, 10, pruned_structure)\n```\n```python\nimport pypaddle.sparse\n\nmodel = pypaddle.sparse.MaskedDeepFFN(784, 10, [100, 100])\n# .. train model\nmodel.generate_structure()  # a networkx graph\n```\n',
    'author': 'Julian Stier',
    'author_email': 'julian.stier@uni-passau.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/innvariant/deepstruct',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)



import math
import os

from enum import Enum

import numpy as np
import torch
import torch.nn as nn

from torch.autograd import grad

import deepstruct.learning
import deepstruct.sparse
import deepstruct.util


class PruneNeuralNetMethod:
    """
    Strategy pattern for the selection of the currently used pruning method.
    Methods can be set during creation of the pruning method object.
    Valid methods are:

    <ul>
        <li>Random Pruning</li>
        <li>Magnitude Pruning Blinded</li>
        <li>Magnitude Pruning Uniform</li>
        <li>Optimal Brain Damage</li>
    </ul>

    The following algorithms are currently under construction:
    <ul>
        <li>Layer-wise Optimal Brain Surgeon</li>
    </ul>

    Method that were considered but due to inefficiency not implemented:
    <ul>
        <li>Optimal Brain Surgeon</li>
        <li>Net-Trim</li>
    </ul>

    All methods except of the random pruning and magnitude based pruning require the loss argument. In order to
    calculate the weight saliency in a top-down approach.
    If no Strategy is specified random pruning will be used as a fallback.
    """

    def __init__(self, method):
        """
        Creates a new PruneNeuralNetMethod object. There are a number of pruning methods supported.

        :param method:      The selected strategy for pruning If no pruning strategy is provided random pruning will be
                            selected as the standard pruning method.
        """
        self.prune_method = method

        # dataset and loss function for error calculation
        self.criterion = None
        self.valid_dataset = None

    def prune(self, network, value):
        """
        Wrapper method which calls the actual pruning strategy and computes how long it takes to complete the pruning
        step.

        :param network:     The network that should be pruned
        :param value:       The percentage of elements that should be pruned
        """
        self.prune_method(self, network, value)

    def requires_loss(self):
        """
        Check if the current pruning method needs the network's loss as an argument.
        :return: True iff a gradient of the network is required.
        """
        return self.prune_method in [
            optimal_brain_damage,
            optimal_brain_damage_absolute,
            optimal_brain_damage_bucket,
            optimal_brain_surgeon_layer_wise,
            optimal_brain_surgeon_layer_wise_bucket,
        ]

    def require_retraining(self):
        """
        Check if the current pruning strategy requires a retraining after the pruning is done
        :return: True iff the retraining is required.
        """
        # todo: does obs-l need retraining?
        # return self.prune_method not in [optimal_brain_surgeon_layer_wise, optimal_brain_surgeon_layer_wise_bucket]
        return True


#
# Top-Down Pruning Approaches
#


def optimal_brain_damage(self, network, percentage):
    """
    Implementation of the optimal brain damage algorithm.
    Requires the gradient to be set in the network.

    :param self:        The strategy pattern object for the pruning method.
    :param network:     The network where the calculations should be done.
    :param percentage:  The percentage of weights that should be pruned.
    """
    # calculate the saliencies for the weights
    calculate_obd_saliency(self, network)
    # prune the elements with the lowest saliency in the network
    prune_network_by_saliency(network, percentage)


def optimal_brain_damage_absolute(self, network, number):
    calculate_obd_saliency(self, network)
    prune_network_by_saliency(network, number, strategy=PruningStrategy.ABSOLUTE)


def optimal_brain_damage_bucket(self, network, bucket_size):
    calculate_obd_saliency(self, network)
    prune_network_by_saliency(network, bucket_size, strategy=PruningStrategy.BUCKET)


#
# Layer-wise approaches
#
def optimal_brain_surgeon_layer_wise(self, network, percentage):
    """
    Layer-wise calculation of the inverse of the hessian matrix. Then the weights are ranked similar to the original
    optimal brian surgeon algorithm.

    :param network:     The network that should be pruned.
    :param percentage:  What percentage of the weights should be pruned.
    :param self:        The strategy pattern object the method is attached to.
    """
    hessian_inverse_path = calculate_obsl_saliency(self, network)

    # prune the elements from the matrix
    for name, layer in deepstruct.sparse.prunable_layers_with_name(network):
        edge_cut(layer, hessian_inverse_path + name + ".npy", value=percentage)


def optimal_brain_surgeon_layer_wise_bucket(self, network, bucket_size):
    hessian_inverse_path = calculate_obsl_saliency(self, network)
    for name, layer in deepstruct.sparse.prunable_layers_with_name(network):
        edge_cut(
            layer,
            hessian_inverse_path + name + ".npy",
            value=bucket_size,
            strategy=PruningStrategy.BUCKET,
        )


#
# Random pruning
#
def random_pruning(self, network, percentage):
    set_random_saliency(network)
    # prune the percentage% weights with the smallest random saliency
    prune_network_by_saliency(network, percentage)


def random_pruning_absolute(self, network, number):
    set_random_saliency(network)
    prune_network_by_saliency(network, number, strategy=PruningStrategy.ABSOLUTE)


#
# Magnitude based approaches
#
def magnitude_class_blinded(self, network, percentage):
    """
    Implementation of weight based pruning. In each step the percentage of not yet pruned weights will be eliminated
    starting with the smallest element in the network.

    The here used method is the class blinded method mentioned in the paper by See et.al from 2016 (DOI: 1606.09274v1).
    The method is also known from the paper by Bachor et.al from 2018 where it was named the PruNet pruning technique
    (DOI: 10.1109/IJCNN.2018.8489764)

    :param network:     The network where the pruning should be done.
    :param percentage:  The percentage of not yet pruned weights that should be deleted.
    """
    prune_network_by_saliency(network, percentage)


def magnitude_class_blinded_absolute(self, network, number):
    prune_network_by_saliency(network, number, strategy=PruningStrategy.ABSOLUTE)


def magnitude_class_uniform(self, network, percentage):
    prune_layer_by_saliency(network, percentage)


def magnitude_class_uniform_absolute(self, network, number):
    prune_layer_by_saliency(network, number, strategy=PruningStrategy.ABSOLUTE)


def magnitude_class_distributed(self, network, percentage):
    """
    This idea comes from the paper 'Learning both Weights and Connections for Efficient Neural Networks'
    (arXiv:1506.02626v3). The main idea is that in each layer respectively to the standard derivation many elements
    should be deleted.
    For each layer prune the weights w for which the following holds:

    std(layer weights) * t > w      This is equal to the following
    t > w/std(layer_weights)        Since std is e(x - e(x))^2 and as square number positive.

    So all elements for which the wright divided by the std. derivation is smaller than some threshold will get deleted

    :param network:     The network that should be pruned.
    :param percentage:  The number of elements that should be pruned.
    :return:
    """
    # set saliency
    set_distributed_saliency(network)
    # prune network
    prune_network_by_saliency(network, percentage)


def magnitude_class_distributed_absolute(self, network, number):
    set_distributed_saliency(network)
    prune_network_by_saliency(network, number, strategy=PruningStrategy.ABSOLUTE)


class PruningStrategy(Enum):
    """
    Enum to represent the different prunuing strategies that can be used.
    Note: not every pruning strategy can be used with every pruning method.
    """

    PERCENTAGE = 0
    ABSOLUTE = 1
    BUCKET = 2


def prune_network_by_saliency(network, value, strategy=PruningStrategy.PERCENTAGE):
    """
    Prune the number of percentage weights from the network. The elements are pruned according to the saliency that is
    set in the network. By default the saliency is the actual weight of the connections. The number of elements are
    pruned blinded. Meaning in each layer a different percentage of elements might get pruned but overall the given one
    is removed from the network. This is a different approach than used in the method below where we prune exactly the
    given percentage from each layer.

    :param network:     The layers of the network.
    :param value:       The number/percentage of elements that should be pruned.
    :param strategy:  If percentage pruning or number of elements pruning is used.
    """
    # calculate the network's threshold
    th = find_network_threshold(network, value, strategy=strategy)

    # set the mask
    for layer in deepstruct.sparse.prunable_layers(network):
        # All deleted weights should be set to zero so they should definetly be less than the threshold since this is
        # positive.
        layer.set_mask(torch.ge(layer.get_saliency(), th).float() * layer.get_mask())


def prune_layer_by_saliency(network, value, strategy=PruningStrategy.PERCENTAGE):
    pre_pruned_weight_count = get_network_weight_count(network).item()

    for layer in deepstruct.sparse.prunable_layers(network):
        mask = list(layer.get_mask().abs().numpy().flatten())
        saliency = list(layer.get_saliency().numpy().flatten())
        _, filtered_saliency = zip(
            *(
                (masked_val, weight_val)
                for masked_val, weight_val in zip(mask, saliency)
                if masked_val == 1
            )
        )

        # calculate threshold
        # percentage pruning
        if strategy is PruningStrategy.PERCENTAGE:
            th = np.percentile(np.array(filtered_saliency), value)
        # absolute pruning
        elif strategy is PruningStrategy.ABSOLUTE:
            # due to floating point operations this is not 100 percent exact a few more or less weights might get
            # deleted
            add_val = round(
                (layer.get_weight_count() / pre_pruned_weight_count * value).item()
            )

            # check if there are enough elements to prune select the highest element and add some penalty to it so the
            if add_val > layer.get_weight_count():
                th = np.argmax(filtered_saliency).item() + 1
            else:
                index = np.argsort(np.array(filtered_saliency))[add_val]
                th = np.array(filtered_saliency)[index].item()
        else:
            raise ValueError("Action is not supported!!!")

        # set mask
        layer.set_mask(torch.ge(layer.get_saliency(), th).float() * layer.get_mask())


def calculate_obd_saliency(self, network):
    # the loss of the network on the cross validation set
    loss = deepstruct.learning.cross_validation_error(
        self.valid_dataset, network, self.criterion
    )

    # calculate the first order gradients for all weights from the pruning layers.
    weight_params = map(
        lambda x: x.get_weight(), deepstruct.sparse.prunable_layers(network)
    )
    loss_grads = grad(loss, weight_params, create_graph=True)

    # iterate over all layers and zip them with their corrosponding first gradient
    for grd, layer in zip(loss_grads, deepstruct.sparse.prunable_layers(network)):
        all_grads = []
        mask = layer.get_mask().view(-1)
        weight = layer.get_weight()

        if torch.cuda.is_available():
            weight.cuda()

        # zip gradient and mask of the network in a lineared fashion
        for num, (g, m) in enumerate(zip(grd.view(-1), mask)):
            if torch.cuda.is_available():
                g.cuda()

            if m.item() == 0:
                # if the element is pruned i.e. if mask == 0 then the second order derivative should e zero as well
                # so no computations are needed
                all_grads += [0]
            else:
                # create the second order derivative and add it to the list which contains all gradients
                drv = grad(g, weight, retain_graph=True)
                all_grads += [drv[0].view(-1)[num].item()]

        # rearrange calculated value to their normal form and set saliency
        layer.set_saliency(
            torch.tensor(all_grads).view(weight.size())
            * layer.get_weight().data.pow(2)
            * 0.5
        )


def calculate_obsl_saliency(self, network):
    out_dir = "./out/hessian"
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)
        os.mkdir(out_dir + "/layerinput")
        os.mkdir(out_dir + "/inverse")

    # where to put the cached layer inputs
    layer_input_path = out_dir + "/layerinput/"
    # where to save the hessian matricies
    hessian_inverse_path = out_dir + "/inverse/"

    # generate the input in the layers and save it for every batch
    keep_input_layerwise(network)

    for i, (images, labels) in enumerate(self.valid_dataset):
        images = images.reshape(-1, 28 * 28)
        network(images)
        for name, layer in deepstruct.sparse.prunable_layers_with_name(network):
            layer_input = layer.layer_input.data.numpy()
            path = layer_input_path + name + "/"
            if not os.path.exists(path):
                os.mkdir(path)

            np.save(path + "layerinput-" + str(i), layer_input)

    # generate the hessian matrix for each layer
    for name, layer in deepstruct.sparse.prunable_layers_with_name(network):
        hessian_inverse_location = hessian_inverse_path + name
        deepstruct.util.generate_hessian_inverse_fc(
            layer, hessian_inverse_location, layer_input_path + name
        )

    return hessian_inverse_path


def edge_cut(
    layer,
    hessian_inverse_path,
    value,
    strategy=PruningStrategy.PERCENTAGE,
):
    """
    This function prune weights of biases based on given hessian inverse and cut ratio
    :param hessian_inverse_path:
    :param layer:
    :param value: The zeros percentage of weights and biases, or, 1 - compression ratio
    :param strategy:
    :return:
    """
    # dataset_size = layer2_input_train.shape[0]
    w_layer = layer.get_weight().data.numpy().T

    # biases = layer.bias.data.numpy()
    n_hidden_1 = w_layer.shape[0]
    n_hidden_2 = w_layer.shape[1]

    sensitivity = np.array([])

    hessian_inverse = np.load(hessian_inverse_path)

    gate_w = layer.get_mask().data.numpy().T
    # gate_b = np.ones([n_hidden_2])

    # calculate number of pruneable elements
    if strategy is PruningStrategy.PERCENTAGE:
        cut_ratio = value / 100  # transfer percentage from full value to floating point
        max_pruned_num = math.floor(layer.get_weight_count() * cut_ratio)
    elif strategy is PruningStrategy.BUCKET:
        max_pruned_num = value
    else:
        raise ValueError("Currently not implemented")

    # Calculate sensitivity score. Refer to Eq.5.
    for i in range(n_hidden_2):
        sensitivity = np.hstack(
            (sensitivity, 0.5 * ((w_layer.T[i] ** 2) / np.diag(hessian_inverse)))
        )
    sorted_index = np.argsort(sensitivity)
    hessian_inverseT = hessian_inverse.T

    prune_count = 0
    for i in range(n_hidden_1 * n_hidden_2):
        prune_index = [sorted_index[i]]
        x_index = math.floor(prune_index[0] / n_hidden_1)  # next layer num
        y_index = prune_index[0] % n_hidden_1  # this layer num

        if gate_w[y_index][x_index] == 1:
            delta_w = (
                -w_layer[y_index][x_index] / hessian_inverse[y_index][y_index]
            ) * hessian_inverseT[y_index]
            gate_w[y_index][x_index] = 0

            if strategy is strategy.PERCENTAGE:
                prune_count += 1
            elif strategy is strategy.BUCKET:
                prune_count += sensitivity[
                    prune_index
                ]  # todo: evaluate here probably a bit wrong :(

            # Parameters update, refer to Eq.5
            w_layer.T[x_index] = w_layer.T[x_index] + delta_w
            # b_layer[x_index] = b_layer[x_index] + delta_w[-1]

        w_layer = w_layer * gate_w
        # b_layer = b_layer * gate_b

        if prune_count == max_pruned_num:
            break

    # set created mask to network again and update the weights
    layer.set_mask(torch.from_numpy(gate_w.T))
    layer.weight = torch.nn.Parameter(torch.from_numpy(w_layer.T))

    # if not os.path.exists(prune_save_path):
    #    os.makedirs(prune_save_path)

    # np.save("%s/weights" % prune_save_path, w_layer)
    # np.save("%s/biases" % prune_save_path, b_layer)


def find_network_threshold(network, value, strategy):
    all_sal = []
    for layer in deepstruct.sparse.prunable_layers(network):
        # flatten both weights and mask
        mask = list(layer.get_mask().abs().numpy().flatten())
        saliency = list(layer.get_saliency().numpy().flatten())

        # zip, filter, unzip the two lists
        _, filtered_saliency = zip(
            *(
                (masked_val, weight_val)
                for masked_val, weight_val in zip(mask, saliency)
                if masked_val == 1
            )
        )
        # add all saliencies to list
        all_sal += filtered_saliency

    # calculate percentile
    if strategy is PruningStrategy.PERCENTAGE:
        return np.percentile(np.array(all_sal), value)
    elif strategy is PruningStrategy.ABSOLUTE:
        # check if there are enough elements to prune
        if value >= len(all_sal):
            return np.argmax(np.array(all_sal)).item() + 1
        else:
            # determine threshold
            index = np.argsort(np.array(all_sal))[value]
            return np.array(all_sal)[index].item()
    elif strategy is PruningStrategy.BUCKET:
        sorted_array = np.sort(all_sal)
        sum_array = 0

        # sum up the elements from the sorted array
        for i in sorted_array:
            sum_array += i
            if sum_array > value:
                # return the last element that has been added to the array
                return i

        # return last element if bucket not reached
        return sorted_array[-1] + 1


def set_random_saliency(network):
    # set saliency to random values
    for layer in deepstruct.sparse.prunable_layers(network):
        layer.set_saliency(torch.rand_like(layer.get_weight()) * layer.get_mask())


def set_random_masks(module: nn.Module):
    if isinstance(module, deepstruct.sparse.MaskedLinearLayer):
        module.set_mask(torch.round(torch.rand_like(module.get_weight())))


def set_distributed_saliency(network):
    # prune from each layer the according number of elements
    for layer in deepstruct.sparse.prunable_layers(network):
        # calculate standard deviation for the layer
        w = layer.get_weight().data
        st_v = 1 / w.std()
        # set the saliency in the layer = weight/st.deviation
        layer.set_saliency(st_v * w.abs())


def reset_pruned_network(network):
    for layer in deepstruct.sparse.prunable_layers(network):
        layer.reset_parameters(keep_mask=True)


def keep_input_layerwise(network):
    for layer in deepstruct.sparse.prunable_layers(network):
        layer.keep_layer_input = True


def get_network_weight_count(network):
    total_weights = 0
    for layer in deepstruct.sparse.prunable_layers(network):
        total_weights += layer.get_weight_count()
    return total_weights
